#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from copy import deepcopy

from paleomix.atomiccmd.builder import \
    apply_options
from paleomix.nodes.samtools import \
    TabixIndexNode, \
    FastaIndexNode, \
    BAMIndexNode
from paleomix.nodes.bedtools import \
    PaddedBedNode
from paleomix.nodes.sequences import \
    ExtractReferenceNode
from paleomix.common.fileutils import \
    swap_ext, \
    add_postfix
from paleomix.nodes.commands import \
    VCFPileupNode, \
    VCFFilterNode, \
    BuildRegionsNode, \
    SampleRegionsNode, \
    GenotypeRegionsNode


def apply_samtools_options(builder, options, argument):
    for (key, value) in dict(options).iteritems():
        sam_argument = key
        if value is not None:
            sam_argument = "%s=%s" % (key, value)

        builder.add_option(argument, sam_argument, sep="=")


###############################################################################
###############################################################################

# Caches for nodes shared between multiple tasks
_BAI_CACHE = {}
_FAI_CACHE = {}
_BED_CACHE = {}
_VCF_CACHE = {}


def build_bam_index_node(bamfile):
    """Returns a node generating a BAI index (using SAMTools) for a BAM file;
    the result is cached, to ensure that multiple calls for the same BAM does
    not result in files being clobbered.

    """
    if bamfile not in _BAI_CACHE:
        _BAI_CACHE[bamfile] = \
            BAMIndexNode(infile=bamfile)
    return _BAI_CACHE[bamfile]


def build_fasta_index_node(reference):
    if reference not in _FAI_CACHE:
        _FAI_CACHE[reference] = \
            FastaIndexNode(infile=reference)
    return _FAI_CACHE[reference]


def build_regions_nodes(regions, padding, dependencies=()):
    destination = add_postfix(regions["BED"], ".padded_%ibp" % (padding,))

    if not padding:
        return regions["BED"], dependencies

    if destination not in _BED_CACHE:
        dependencies = list(dependencies)
        dependencies.append(build_fasta_index_node(regions["FASTA"]))
        _BED_CACHE[destination] \
            = PaddedBedNode(fai_file=regions["FASTA"] + ".fai",
                            infile=regions["BED"],
                            outfile=destination,
                            amount=padding,
                            dependencies=dependencies)

    return destination, (_BED_CACHE[destination],)


def _apply_vcf_filter_options(vcffilter, genotyping, sample):
    filter_cfg = genotyping["VCF_Filter"]
    apply_options(vcffilter.commands["filter"], filter_cfg)
    if filter_cfg["MaxReadDepth"][sample]:
        max_depth = filter_cfg["MaxReadDepth"][sample]
        vcffilter.commands["filter"].set_option("--max-read-depth", max_depth)
    return vcffilter.build_node()


def build_genotyping_bedfile_nodes(options, genotyping, sample, regions,
                                   dependencies):
    bamfile = "%s.%s.bam" % (sample, regions["Prefix"])
    bamfile = os.path.join(options.samples_root, bamfile)
    if regions["Realigned"]:
        bamfile = add_postfix(bamfile, ".realigned")

    prefix = regions["Genotypes"][sample]
    padding, bedfile = genotyping["Padding"], None
    if not genotyping["GenotypeEntirePrefix"]:
        bedfile, nodes = \
            build_regions_nodes(regions, padding, dependencies)
        bai_node = build_bam_index_node(bamfile)
        dependencies = nodes + (bai_node,)
    else:
        prefix = os.path.join(os.path.dirname(prefix),
                              "%s.%s.TEMP" % (sample, regions["Prefix"]))

        dependencies += (build_bam_index_node(bamfile),)

    return prefix, bamfile, bedfile, dependencies


def build_genotyping_nodes_cached(options, genotyping, sample, regions,
                                  dependencies):
    """Carries out genotyping, filtering of calls, and indexing of files for a
    given sample and prefix. If the option 'GenotypeEntirePrefix' is enabled,
    the BAM is genotyped once, and each set of RegionsOfInterest simply extract
    the relevant regions during construction of the consensus sequence.

    Parameters:
        options: An options object (c.f. paleomix.tools.phylo_pipeline.config).
        genotyping: Genotyping options defined for a specific set of areas of
                    interest, corresponding to Genotyping:NAME in the makefile.
        sample: The name of the sample to be genotyped.
        egions: A dictionary for a 'RegionsOfInterest' from the makefile.
        dependencies: Depenencies that must be met before genotyping starts.

    Returns a tuple containing the filename of the filtered and tabix-indexed
    VCF file, and the top-level node generating this file. Multiple calls for
    the same BAM and prefix will return the same VCF and nodes if the option
    for 'GenotypeEntirePrefix' is enabled, otherwise each ROI is genotyped
    individiually.

    Output files are generated in ./results/PROJECT/genotyping. If the option
    for 'GenotypeEntirePrefix' is enabled, the following files are generated:
        SAMPLE.PREFIX.vcf.bgz: Unfiltered calls for variant/non-variant sites.
        SAMPLE.PREFIX.vcf.pileup.bgz: Pileup of sites containing SNPs.
        SAMPLE.PREFIX.vcf.pileup.bgz.tbi: Tabix index of the pileup.
        SAMPLE.PREFIX.filtered.vcf.bgz: Variant calls filtered with vcf_filter.
        SAMPLE.PREFIX.filtered.vcf.bgz.tbi: Tabix index for the filtered VCF.

    If 'GenotypeEntirePrefix' is not enabled for a given ROI, the following
    files are generated for that ROI (see descriptions above):
        SAMPLE.PREFIX.ROI.filtered.vcf.bgz
        SAMPLE.PREFIX.ROI.filtered.vcf.bgz.tbi
        SAMPLE.PREFIX.ROI.vcf.bgz
        SAMPLE.PREFIX.ROI.vcf.pileup.bgz
        SAMPLE.PREFIX.ROI.vcf.pileup.bgz.tbi

    In addition, the following files are generated for each set of
    RegionsOfInterest (ROI), regardless of the 'GenotypeEntirePrefix' option:
        SAMPLE.PREFIX.ROI.CDS.fasta: FASTA sequence of each feature in the ROI.
        SAMPLE.PREFIX.ROI.CDS.fasta.fai: FASTA index generated using SAMTools.

    """
    output_prefix, bamfile, bedfile, dependencies \
        = build_genotyping_bedfile_nodes(options, genotyping, sample, regions,
                                         dependencies)

    if (bamfile, output_prefix) in _VCF_CACHE:
        return _VCF_CACHE[(bamfile, output_prefix)]

    calls = swap_ext(output_prefix, ".vcf.bgz")
    pileups = swap_ext(output_prefix, ".vcf.pileup.bgz")
    filtered = swap_ext(output_prefix, ".filtered.vcf.bgz")

    # 1. Call samtools mpilup | bcftools view on the bam
    genotype = GenotypeRegionsNode.customize(reference=regions["FASTA"],
                                             bedfile=bedfile,
                                             infile=bamfile,
                                             outfile=calls,
                                             nbatches=options.samtools_max_threads,
                                             dependencies=dependencies)

    apply_samtools_options(genotype.command, genotyping["MPileup"],
                           "--mpileup-argument")
    apply_samtools_options(genotype.command, genotyping["BCFTools"],
                           "--bcftools-argument")
    genotype = genotype.build_node()

    # 2. Collect pileups of sites with SNPs, to allow proper filtering by
    #    frequency of the minor allele, as only the major non-ref allele is
    #    counted in the VCF (c.f. field DP4).
    vcfpileup = VCFPileupNode.customize(reference=regions["FASTA"],
                                        infile_bam=bamfile,
                                        infile_vcf=calls,
                                        outfile=pileups,
                                        dependencies=genotype)
    apply_samtools_options(vcfpileup.command, genotyping["MPileup"],
                           "--mpileup-argument")
    vcfpileup = vcfpileup.build_node()

    vcf_tabix = TabixIndexNode(infile=pileups,
                               preset="pileup",
                               dependencies=vcfpileup)

    # 3. Filter all sites using the 'vcf_filter' command
    vcffilter = VCFFilterNode.customize(infile=calls,
                                        pileup=pileups,
                                        outfile=filtered,
                                        regions=regions,
                                        dependencies=vcf_tabix)
    vcffilter = _apply_vcf_filter_options(vcffilter, genotyping, sample)

    # 4. Tabix index. This allows random-access to the VCF file when building
    #    the consensus FASTA sequence later in the pipeline.
    tabix = TabixIndexNode(infile=filtered,
                           preset="vcf",
                           dependencies=vcffilter)

    _VCF_CACHE[(bamfile, output_prefix)] = (filtered, tabix)
    return filtered, tabix


def build_genotyping_nodes(options, genotyping, sample, regions, dependencies):
    """Builds the nodes required for genotyping a BAM, in part or in whole.

    By default, only the region of interest (including padding) will be
    genotyped. However, if option 'GenotypeEntirePrefix' is enabled, the entire
    genome is genotyped, and reused between different areas of interest.

    In addition to the files generated by 'build_genotyping_nodes_cached', this
    function generates the following files:
        SAMPLE.PREFIX.ROI.fasta: FASTA containing each named region.
        SAMPLE.PREFIX.ROI.fasta.fai: Index file built using "samtools faidx"

    The function returns a sequence of the top-level nodes generating the files.

    """
    # 1. Get path of the filtered VCF file, and the assosiated node
    filtered, node = build_genotyping_nodes_cached(options=options,
                                                   genotyping=genotyping,
                                                   sample=sample,
                                                   regions=regions,
                                                   dependencies=dependencies)

    # 2. Generate consensus sequence from filtered VCF
    output_fasta = regions["Genotypes"][sample]
    builder = BuildRegionsNode.customize(infile=filtered,
                                         bedfile=regions["BED"],
                                         outfile=output_fasta,
                                         padding=genotyping["Padding"],
                                         dependencies=node)
    if regions["ProteinCoding"]:
        builder.command.set_option("--whole-codon-indels-only")
    if not regions["IncludeIndels"]:
        builder.command.set_option("--ignore-indels")
    builder = builder.build_node()

    # 3. Index sequences to make retrival easier for MSA
    faidx = FastaIndexNode(infile=output_fasta,
                           dependencies=builder)

    return (faidx,)


def build_sampling_nodes(options, genotyping, sample, regions, dependencies):
    fasta_file = regions["Genotypes"][sample]
    pileup_file = swap_ext(fasta_file, ".pileup.bgz")

    padding = genotyping["Padding"]
    slop, node = build_regions_nodes(regions, padding, dependencies)

    bam_file = "%s.%s.bam" % (sample, regions["Prefix"])
    bam_file = os.path.join(options.samples_root, bam_file)
    if regions["Realigned"]:
        bam_file = add_postfix(bam_file, ".realigned")
    bai_node = build_bam_index_node(bam_file)

    genotype = GenotypeRegionsNode.customize(pileup_only=True,
                                             reference=regions["FASTA"],
                                             bedfile=slop,
                                             infile=bam_file,
                                             outfile=pileup_file,
                                             nbatches=options.samtools_max_threads,
                                             dependencies=node + (bai_node,))
    apply_samtools_options(genotype.command, genotyping["MPileup"],
                           "--mpileup-argument")
    genotype = genotype.build_node()

    tabix = TabixIndexNode(infile=pileup_file,
                           preset="pileup",
                           dependencies=genotype)

    builder = SampleRegionsNode(infile=pileup_file,
                                bedfile=regions["BED"],
                                outfile=fasta_file,
                                dependencies=tabix)

    faidx = FastaIndexNode(infile=fasta_file,
                           dependencies=builder)

    return (faidx,)


def build_reference_nodes(options, genotyping, sample, regions, dependencies):
    input_file = "%s.%s.fasta" % (regions["Prefix"], sample)
    input_fpath = os.path.join(options.refseq_root, input_file)

    output_file = "%s.%s.fasta" % (sample, regions["Desc"])
    output_fpath = os.path.join(options.destination, "genotypes", output_file)

    dependencies = list(dependencies)
    dependencies.append(build_fasta_index_node(regions["FASTA"]))

    node = ExtractReferenceNode(reference=input_fpath,
                                bedfile=regions["BED"],
                                outfile=output_fpath,
                                dependencies=dependencies)

    faidx = FastaIndexNode(infile=output_fpath,
                           dependencies=node)
    return (faidx,)


# Functions used to carry out each of the supported genotyping methods
_GENOTYPING_METHODS = {
    "reference sequence": build_reference_nodes,
    "random sampling": build_sampling_nodes,
    "samtools": build_genotyping_nodes,
}


def build_sample_nodes(options, genotyping, regions_sets, sample,
                       dependencies=()):
    nodes = []
    for regions in regions_sets.itervalues():
        regions = deepcopy(regions)

        # Enforce homozygous contigs based on gender tag
        regions["HomozygousContigs"] \
            = regions["HomozygousContigs"][sample["Gender"]]

        genotyping_method = sample["GenotypingMethod"].lower()
        if genotyping_method not in _GENOTYPING_METHODS:
            assert False, "Unexpected genotyping method %r for sample %r" \
                          % (genotyping_method, sample["Name"])

        genotyping_function = _GENOTYPING_METHODS[genotyping_method]
        node = genotyping_function(options=options,
                                   genotyping=genotyping[regions["Name"]],
                                   sample=sample["Name"],
                                   regions=regions,
                                   dependencies=dependencies)
        nodes.extend(node)

    return nodes


def chain(pipeline, options, makefiles):
    destination = options.destination
    for makefile in makefiles:
        regions_sets = makefile["Project"]["Regions"]
        genotyping = makefile["Genotyping"]
        options.destination = os.path.join(destination,
                                           makefile["Project"]["Title"])

        nodes = []
        for sample in makefile["Project"]["Samples"].itervalues():
            nodes.extend(build_sample_nodes(options, genotyping, regions_sets,
                                            sample, makefile["Nodes"]))

        makefile["Nodes"] = tuple(nodes)
    options.destination = destination



#!/usr/bin/python
#
# Copyright (c) 2016 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import os

import pysam

import paleomix
import paleomix.resources

from paleomix.node import Node

import paleomix.common.fileutils as fileutils

import paleomix.tools.zonkey.parts.admixture as admixture
import paleomix.tools.zonkey.parts.nuclear as nuclear

from paleomix.tools.zonkey.common import \
    RSCRIPT_VERSION, \
    read_summary

from paleomix.nodes.samtools import SAMTOOLS_VERSION
from paleomix.nodes.raxml import RAXML_VERSION


class ReportNode(Node):
    def __init__(self, config, root, has_nuc, has_mt, dependencies=()):
        """

        Arguments:
          config -- Config object generated using paleomix.tools.zonkey.config.
          root -- Root folder containing current analysis.
          has_nuc -- True if a nuclear BAM was provided.
          has_mt -- True if a mitochondrial BAM was provided.
          dependencies -- Nodes for ReportNode to depend on.
        """

        self._root = root
        self._data = copy.deepcopy(config.database)
        self._report = AnalysisReport(config, root, has_nuc, has_mt)
        self._has_nuc = bool(has_nuc)
        self._has_mt = bool(has_mt)

        self._treemix_outgroup = config.treemix_outgroup
        self._treemix_k = config.treemix_k
        if self._treemix_k is None:
            self._treemix_k = '<automatic>'

        Node.__init__(self,
                      description="<Report -> %r>"
                      % (os.path.join(self._root, "report.html"),),
                      input_files=self._report.input_files(),
                      output_files=(os.path.join(self._root, "report.html"),
                                    os.path.join(self._root, "report.css")),
                      dependencies=dependencies)

    def _run(self, _config, temp):
        with open(os.path.join(temp, "report.html"), "w") as output_handle:
            header = _HTML_HEADER.format(Version=paleomix.__version__,
                                         Sidebar=self._build_sidebar())
            output_handle.write(header)

            self._write_overview(output_handle)
            self._write_sample_description(output_handle)

            if self._has_nuc:
                self._write_admixture_estimates(output_handle)
                self._write_pca_plots(output_handle)
                self._write_treemix_plots(output_handle)

            if self._has_mt:
                self._write_mitochondrial_plots(output_handle)

            self._write_references(output_handle)

            output_handle.write(_HTML_FOOTER)

    def _teardown(self, config, temp):
        fileutils.make_dirs(self._root)

        fileutils.move_file(os.path.join(temp, "report.html"),
                            os.path.join(self._root, "report.html"))

        css_path = paleomix.resources.report("zonkey", "report.css")
        fileutils.copy_file(css_path, os.path.join(self._root, "report.css"))

    def _write_overview(self, output_handle):
        output_handle.write(_SECTION_HEADER.format(name="overview",
                                                   title="Overview"))

        overview = _OVERVIEW_HEADER.format(PYSAM=pysam.__version__,
                                           SAMTOOLS=_fmt_v(SAMTOOLS_VERSION),
                                           PLINK=_fmt_v(nuclear.PLINK_VERSION),
                                           RSCRIPT=_fmt_v(RSCRIPT_VERSION))
        output_handle.write(overview)

        if self._has_nuc:
            summary = self._report.snp_summary()
            output_handle.write(_OVERVIEW_NUCLEAR % summary)

        if self._has_mt:
            summary = self._report.mito_summary()
            output_handle.write(_OVERVIEW_MITOCHONDRIA % summary)

        if self._has_nuc:
            output_handle.write(_OVERVIEW_NUCLEAR_COVERAGE)

        output_handle.write(_OVERVIEW_FOOTER)

    def _write_sample_description(self, output_handle):
        output_handle.write(_SECTION_HEADER.format(name="samples",
                                                   title="Reference Panel"))

        output_handle.write(_SAMPLE_LIST_HEADER)

        last_group_2 = None
        last_group_3 = None
        for row in sorted(self._data.samples.itervalues(),
                          key=lambda row: (row["Group(2)"],
                                           row["Group(3)"],
                                           row["ID"])):

            row = dict(row)
            if last_group_2 != row["Group(2)"]:
                last_group_2 = row["Group(2)"]
                last_group_3 = row["Group(3)"]
            else:
                row["Group(2)"] = ""

                if last_group_3 != row["Group(3)"]:
                    last_group_3 = row["Group(3)"]
                else:
                    row["Group(3)"] = ""

            pub = row["Publication"]
            if pub.startswith("http"):
                row["Publication"] \
                    = '<a href="{0}">Link</a>'.format(pub.strip())
            elif row["Publication"].startswith("doi:"):
                pub = pub[4:].strip()
                url = "https://doi.org/{}".format(pub)
                row["Publication"] \
                    = 'doi:<a href="{0}">{1}</a>'.format(url, pub)

            output_handle.write(_SAMPLE_LIST_ROW.format(**row))

        output_handle.write("      </table>\n")

    def _write_admixture_estimates(self, output_handle):
        header = _SECTION_HEADER.format(name="admixture",
                                        title="Admixture Estimates")
        output_handle.write(header)

        admixture_v = _fmt_v(nuclear.ADMIXTURE_VERSION)
        overview = _ADMIXTURE_OVERVIEW.format(ADMIXTURE=admixture_v)
        output_handle.write(overview)

        for k_groups in (2, 3):
            summary_incl = self._build_admixture_cell(k_groups, True)
            summary_excl = self._build_admixture_cell(k_groups, False)

            output_handle.write(_ADMIXTURE_ROW.format(K=k_groups,
                                                      Incl_TS=summary_incl,
                                                      Excl_TS=summary_excl))

    def _build_admixture_cell(self, k_groups, incl_ts,
                              cutoff=admixture.CUTOFF):
        try:
            groups = self._report.admixture_results(k_groups, incl_ts)
        except admixture.AdmixtureError, error:
            return "<strong>ERROR: {}</strong".format(error)

        if sum((value >= cutoff) for _, value in groups) < 2:
            return "<strong>No admixture detected.</strong>"

        lines = [
            "<strong>Possible hybridization detected:</strong>",
            "<ul>",
        ]

        for group, value in groups:
            if value >= cutoff:
                name = " / ".join(sorted(group))

                lines.append("  <li>%s (%.2f%%)</li>" % (name, value * 100))

        lines.append("</ul>")

        if sum((value >= cutoff) for _, value in groups) != 2:
            return "\n            ".join(lines)

        percentiles = self._report.admixture_percentiles(data=self._data,
                                                         k_groups=k_groups,
                                                         incl_ts_k=incl_ts)

        if not ('Lower' in percentiles or 'Upper' in percentiles):
            lines.append("WARNING: Could not determine percentiles.")

            return "\n            ".join(lines)

        if 'Lower' not in percentiles:
            percentiles['Lower'] = percentiles['Upper']
            finale = \
                '; note that this is more simulated reads than what was ' \
                'processed in this analyses, potentially resulting in an ' \
                'overrestimating of percentages.'
        elif 'Upper' not in percentiles:
            percentiles['Upper'] = percentiles['Lower']
            finale = \
                '; note that this is fewer simulated reads than what was ' \
                'processed in this analyses, potentially resulting in an ' \
                'underrestimating of percentages'
        else:
            finale = '.'

        lower_pct = "%.1f" % ((1.0 - max(percentiles['Lower']['Upper'],
                                         percentiles['Upper']['Lower'])) *
                              100.0,)
        upper_pct = "%.1f" % ((1.0 - min(percentiles['Lower']['Upper'],
                                         percentiles['Upper']['Lower'])) *
                              100.0,)

        pct_range = lower_pct
        if lower_pct != upper_pct:
            pct_range = '%s - %s' % (lower_pct, upper_pct)

        lower_reads = min(percentiles['Lower']['NReads'],
                          percentiles['Upper']['NReads'])
        upper_reads = max(percentiles['Lower']['NReads'],
                          percentiles['Upper']['NReads'])

        reads_range = lower_reads
        if lower_reads != upper_reads:
            reads_range = '%s to %s' % (lower_reads, upper_reads)

        lines.append('Admixture results fall within %s percent of those '
                     'observed for simulated F1 %s / %s hybrids, based on '
                     '%s randomly selected reads%s'
                     % (pct_range,
                        percentiles['Sample1'],
                        percentiles['Sample2'],
                        reads_range, finale))

        return "\n            ".join(lines)

    def _write_pca_plots(self, output_handle):
        output_handle.write(_SECTION_HEADER.format(name="pca",
                                                   title="PCA Plots"))

        smartpca_v = _fmt_v(nuclear.SMARTPCA_VERSION)
        output_handle.write(_PCA_SECTION.format(SMARTPCA=smartpca_v))

    def _write_treemix_plots(self, output_handle):
        output_handle.write(_SECTION_HEADER.format(name="treemix",
                                                   title="Treemix Plots"))

        outgroups = ""
        if self._treemix_outgroup:
            outgroups = ", ".join(map(repr, self._treemix_outgroup))
            outgroups = ". The tree was rooted on the clade containing the " \
                        "sample(s) %s" % (outgroups)

        treemix_v = _fmt_v(nuclear.TREEMIX_VERSION)
        overview = _TREEMIX_OVERVIEW.format(treemix_k=self._treemix_k,
                                            treemix_outgroup=outgroups,
                                            TREEMIX=treemix_v)
        output_handle.write(overview)

        for prefix in ("incl_ts", "excl_ts"):
            output_handle.write("<h2>%s</h2>\n" % (_TS_LABELS[prefix],))

            for n_edges in (0, 1):
                variance_file = os.path.join(self._root,
                                             "figures",
                                             "treemix",
                                             "%s_%i_variance.txt")

                with open(variance_file % (prefix, n_edges)) as handle:
                    variance = handle.read().strip()

                treemix_row = _TREEMIX_TREE_ROW.format(Prefix=prefix,
                                                       Edges=n_edges,
                                                       Variance=variance)
                output_handle.write(treemix_row)

    def _write_mitochondrial_plots(self, output_handle):
        header = _SECTION_HEADER.format(name="mitochondria",
                                        title="Mitochondrial Phylogeny")
        output_handle.write(header)
        raxml_v = _fmt_v(RAXML_VERSION)
        output_handle.write(_MITOCONDRIAL_SECTION.format(RAXML=raxml_v))

    def _write_references(self, output_handle):
        header = _SECTION_HEADER.format(name="references",
                                        title="References")
        output_handle.write(header)
        output_handle.write(_REFERENCES)

    def _build_sidebar(self):
        lines = [_SIDEBAR_HEADER]

        if self._has_nuc:
            lines.append(_SIDEBAR_NUCLEAR)

        if self._has_mt:
            lines.append(_SIDEBAR_MITO)

        lines.append(_SIDEBAR_FOOTER)

        return "\n".join(lines)


class AnalysisReport(object):
    def __init__(self, config, root, has_nuc, has_mt):
        self._has_nuc = bool(has_nuc)
        self._has_mt = bool(has_mt)
        self._filtered = bool(config.indep)
        self._config = config
        self._root = root
        self._data = config.database

    def input_files(self):
        input_files = [self._config.tablefile]
        if self._has_nuc:
            # Summary file generated while building TPED files
            input_files.append(os.path.join(self._root,
                                            "plink",
                                            "common.summary"))

            for postfix in ('incl_ts', 'excl_ts'):
                admix_root = os.path.join(self._root, "admixture")

                if self._filtered:
                    # Required to count number of SNPs included after filtering
                    input_files.append(os.path.join(self._root,
                                                    "plink",
                                                    postfix + ".bim"))

                # Include files showing proproation of ancestral populations,
                # which are required to build admixture figures in the reports.
                for k_groups in (2, 3):
                    input_files.append(os.path.join(admix_root,
                                                    "%s.%i.Q" % (postfix,
                                                                 k_groups)))

                # Include files tabulating variance explained by models
                figures_path = os.path.join(self._root, "figures", "treemix")
                for n_edges in (0, 1):
                    variance_path = os.path.join(figures_path,
                                                 "%s_%i_variance.txt"
                                                 % (postfix, n_edges))

                    input_files.append(variance_path)

        if self._has_mt:
            input_files.append(os.path.join(self._root,
                                            "mitochondria",
                                            "sequences.summary"))

        return input_files

    def snp_summary(self):
        summary = read_summary(os.path.join(self._root, "plink",
                                            "common.summary"))

        if self._filtered:
            for postfix in ('incl_ts', 'excl_ts'):
                key = "n_sites_%s" % (postfix,)

                filename = os.path.join(self._root, "plink", postfix + ".bim")
                with open(filename) as handle:
                    n_used = sum(1 for _ in handle)

                summary[key] = "%i of %i" % (n_used, summary[key])

        return summary

    def mito_summary(self):
        return read_summary(os.path.join(self._root,
                                         "mitochondria",
                                         "sequences.summary"))

    def admixture_results(self, k_groups, incl_ts,
                          cutoff=admixture.CUTOFF):
        prefix = "incl_ts" if incl_ts else "excl_ts"
        filename = os.path.join(self._root, "admixture",
                                "%s.%i.Q" % (prefix, k_groups))

        return admixture.read_admixture_results(filename=filename,
                                                data=self._config.database,
                                                k_groups=k_groups,
                                                cutoff=cutoff)

    def admixture_percentiles(self, data, k_groups, incl_ts_k,
                              cutoff=admixture.CUTOFF):
        try:
            results = self.admixture_results(k_groups, incl_ts_k, cutoff)
        except admixture.AdmixtureError:
            return

        groups = [group for group, value in results if value >= cutoff]
        if len(groups) != 2:
            return

        sample1, = groups[0]
        sample2, = groups[1]
        delta = abs(max(value for _, value in results) - 0.5)
        summary = read_summary(os.path.join(self._root, "plink",
                                            "common.summary"))

        return admixture.get_percentiles(data=data,
                                         sample1=sample1,
                                         sample2=sample2,
                                         nreads=summary['n_reads_used'],
                                         k_groups=k_groups,
                                         has_ts=incl_ts_k,
                                         value=delta)


def _fmt_v(requirement):
    return ".".join(map(str, requirement.version))


###############################################################################

_TS_LABELS = {
    "incl_ts": "Including transitions",
    "excl_ts": "Excluding transitions",
}


###############################################################################

_HTML_HEADER = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<title>PALEOMIX Zonkey v{Version}</title>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<link href="report.css" rel="stylesheet" type="text/css" />
</head>
<body>
<a name="top" id="top"></a>
<center>
  <div id="header">
    <h1>PALEOMIX Zonkey v{Version}</h1>
    <h2>A pipeline for detection of F1 hybrids in equids.</h2>
  </div>
  <div id="content">
    <p class="introduction">
      Schubert M, Ermini L, Sarkissian CD, J&oacute;nsson H, Ginolhac A,
      Schaefer R, Martin MD, Fern&aacute;ndez R, Kircher M, McCue M,
      Willerslev E, and Orlando L. "<strong>Characterization of ancient and
      modern genomes by SNP detection and phylogenomic and metagenomic analysis
      using PALEOMIX</strong>". Nat Protoc. 2014 May;9(5):1056-82. doi:<a
      href="https://doi.org/10.1038/nprot.2014.063">10.1038/nprot.2014.063</a>.
      Epub 2014 Apr 10. PubMed PMID: <a
      href="http://www.ncbi.nlm.nih.gov/pubmed/24722405">24722405</a>.
    </p>
{Sidebar}
    <div id="mainbar">
"""

_SECTION_HEADER = """      <h1><a name="{name}" id="{name}"></a>{title}</h1>
"""

_OVERVIEW_HEADER = """
      <p> Data processed using
        <a href="https://github.com/pysam-developers/pysam">pysam</a> v{PYSAM},
        <a href="https://samtools.github.io/">SAMTools</a> v{SAMTOOLS}
        [<em>Li <em>et al.</em> 2009</em>] and
        <a href="http://pngu.mgh.harvard.edu/purcell/plink/">PLINK</a> v{PLINK}
        [<em>Purcell <em>et al.</em> 2007</em>]; plotting was carried out using
        <a href="https://www.r-project.org/">R</a> v{RSCRIPT}. Additional
        tools listed below.

      <div style="display:table;width:100%;">
          <div style="display:table-cell;">
"""

_OVERVIEW_NUCLEAR = """
            <strong>Nuclear report from '<em>%(filename)s</em>'</strong>

            <table style="width:95%%">
              <tr>
                <td style="width:50%%;">Number of reads processed:</td>
                <td>%(n_reads)s</td>
              </tr>
              <tr>
                <td>Number of reads overlapping SNPs:</td>
                <td>%(n_reads_used)s</td>
              </tr>
              <tr>
                <td>Number of SNPs used (incl. transitions):</td>
                <td>%(n_sites_incl_ts)s</td>
              </tr>
              <tr>
                <td>Number of SNPs used (excl. transitions):</td>
                <td>%(n_sites_excl_ts)s</td>
              </tr>
            </table>
"""

_OVERVIEW_MITOCHONDRIA = """
            <br>

            <h4>Mitochondrial report from '<em>%(filename)s</em>'</h4>

            <table style="width:95%%">
              <tr>
                <td style="width:50%%;">Reference sequence used:</td>
                <td>%(sequence_name)s</td>
              </tr>
              <tr>
                <td>Reference sequence length:</td>
                <td>%(sequence_len)s</td>
              </tr>
              <tr>
                <td>Number of sites covered:</td>
                <td>%(covered_sites)s</td>
              </tr>
              <tr>
                <td>Percentage of sites covered:</td>
                <td>%(covered_pct)s</td>
              </tr>
              <tr>
                <td>Mean coverage per site:</td>
                <td>%(mean_coverage)s</td>
              </tr>
            </table>
"""


_OVERVIEW_NUCLEAR_COVERAGE = """
          </div>
          <div style="display:table-cell;width:350px;">
           <strong>Autosomes vs. sex-chromosomes:</strong>
           <a href="figures/coverage/coverage.pdf">
             <img src="figures/coverage/coverage.png"
                  style="vertical-align:top;horizontal-align:center;">
           </a>
"""

_OVERVIEW_FOOTER = """
          </div>
      </div>
"""

_SAMPLE_LIST_HEADER = """
      <table summary="List of samples in the reference panel.">
        <tr>
          <th>Group(2)</th>
          <th>Group(3)</th>
          <th>ID</th>
          <th>Species</th>
          <th>Sex</th>
          <th>Sample Name</th>
          <th>Publication</th>
        </tr>
"""

_SAMPLE_LIST_ROW = """
        <tr>
          <td>{Group(2)}
          <td>{Group(3)}
          <td>{ID}</td>
          <td><em>{Species}</em></td>
          <td>{Sex}</th>
          <td>{SampleID}</td>
          <td>{Publication}</td>
        </tr>
"""


_ADMIXTURE_OVERVIEW = """
      <p>
        Admixture proportions estimated using
        <a href="https://www.genetics.ucla.edu/software/admixture/">ADMIXTURE</a>
        v{ADMIXTURE} <em>[Alexander <em>et al.</em> 2009]</em>, using default
        parameters.
      </p>
"""


_ADMIXTURE_ROW = """
      <table summary="Admixture between sample and reference populations.">
        <tr>
          <th>{K} ancestral groups</th>
          <th>{K} ancestral groups, excluding transitions</th>
        </tr>
        <tr>
          <td>
            <a href="figures/admixture/incl_ts_k{K}.pdf">
                <img style="width:95%" src="figures/admixture/incl_ts_k{K}.png"
                     alt="Admixture for k={K}, including transitions." />
            </a>
            <br/>
            <br/>
            {Incl_TS}
          </td>
          <td>
            <a href="figures/admixture/excl_ts_k{K}.pdf">
                <img style="width:95%" src="figures/admixture/excl_ts_k{K}.png"
                     alt="Admixture for k={K}, excluding transitions." />
            </a>
            <br/>
            <br/>
            {Excl_TS}
          </td>
        </tr>
      </table>
"""


_PCA_SECTION = """
      <p>
        Principal Component Analysis carried out using SmartPCA v{SMARTPCA},
        from the <a
        href="http://www.hsph.harvard.edu/alkes-price/software/">EIGENSOFT</a>
        toolkit.
      </p>

      <table summary="PCA plots comparing sample with the reference panel.">
        <tr>
          <th>Including transitions</th>
          <th>Excluding transitions</th>
        </tr>
        <tr>
          <td>
            <a href="figures/pca/incl_ts.pdf">
                <img style="width:95%" src="figures/pca/incl_ts.png"
                     alt="PCA plot, including transitions." />
            </a>
          </td>
          <td>
            <a href="figures/pca/excl_ts.pdf">
                <img style="width:95%" src="figures/pca/excl_ts.png"
                     alt="PCA plot, excluding transitions." />
            </a>
          </td>
        </tr>
      </table>
"""


_TREEMIX_OVERVIEW = """
      <p>
        Detection of population mixture using
        <a href="https://bitbucket.org/nygcresearch/treemix/wiki/Home">TreeMix</a>
        v{TREEMIX} <em>[Pickrell and Pritchard 2012]</em>; parameters were -k
        {treemix_k}; -global; and supervised estimation using ancestral groups
        listed in the Reference Panel{treemix_outgroup}.
      </p>
"""


_TREEMIX_TREE_ROW = """
      <table summary="Treemix plots, for {Edges} edge(s).">
        <tr>
          <th>Edges = {Edges}</th>
          <th>Residuals</th>
        </tr>
        <tr>
          <td>
            <a href="figures/treemix/{Prefix}_{Edges}_tree.pdf">
                <img style="width:95%"
                     src="figures/treemix/{Prefix}_{Edges}_tree.png"
                     alt="Treemix plot, {Edges} edge(s), incl. transitions." />
            </a>
          </td>
          <td>
            <a href="figures/treemix/{Prefix}_{Edges}_residuals.pdf">
                <img style="width:95%"
                     src="figures/treemix/{Prefix}_{Edges}_residuals.png"
                     alt="Treemix plot, {Edges} edge(s), excl. transitions." />
            </a>
          </td>
        </tr>
      </table>

      <p>
        Variance explained by model = {Variance}.
      </p>
"""


_MITOCONDRIAL_SECTION = """
      <p>
        Phylogenetic inference performed using RAxML
        v{RAXML} [<em>Stamatakis 2006</em>].
      <p>

      <div style="text-align:center;">
        <a href="figures/mitochondria/mito_phylo.pdf">
          <img src="figures/mitochondria/mito_phylo.png"
               alt="Mitochondrial maximum likelihood phylogeny." />
        </a>
      </div>
"""


_REFERENCES = """
        <p>
          <ul>
            <li>
              Alexander <em>et al</em>. "<strong>Fast model-based estimation of
              ancestry in unrelated individuals</strong>". <em>Genome Res</em>.
              2009 Sep;19(9):1655-64.
                doi:<a href="https://doi.org/10.1101/gr.094052.109">
                    10.1101/gr.094052.109</a>.
                PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/19648217">
                  19648217</a>.
            </li>
            <li>
              Li <em>et al</em>. "<strong>The Sequence Alignment/Map format and
              SAMtools</strong>. <em>Bioinformatics</em>. 2009 Aug
              15;25(16):2078-9.
                doi:<a href="https://doi.org/10.1093/bioinformatics/btp352">
                    10.1093/bioinformatics/btp352</a>.
                PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/19505943">
                  19505943</a>.
            </li>
            <li>
              Pickrell and Pritchard. "<strong>Inference of population splits
              and mixtures from genome-wide allele frequency data</strong>".
              <em>PLoS Genet</em>. 2012;8(11):e1002967.
              doi:<a href="https://doi.org/10.1371/journal.pgen.1002967">
                  10.1371/journal.pgen.1002967</a>.
              PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/23166502">
                23166502</a>.
            </li>
            <li>
              Purcell <em>et al</em>. "<strong>PLINK: a tool set for whole-
              genome association and population-based linkage
              analyses</strong>". <em>Am J Hum Genet</em>. 2007
              Sep;81(3):559-75.
              PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/17701901">
                17701901</a>.
            </li>
            <li>
              Stamatakis. "<strong>RAxML-VI-HPC: maximum likelihood-based
              phylogenetic analyses with thousands of taxa and mixed
              models</strong>". <em>Bioinformatics</em>. 2006 Nov
              1;22(21):2688-90. Epub 2006 Aug 23.
              doi:<a href="https://doi.org/10.1093/bioinformatics/btl446">
                  10.1093/bioinformatics/btl446</a>.
              PMID:<a href="http://www.ncbi.nlm.nih.gov/pubmed/16928733">
                16928733</a>.
          </ul>
        </p>
"""


_HTML_FOOTER = """
    </div>
  </div>
  <div id="footer">
    This report is based on the PLAIN 1.0 design by
    <a href="http://www.sixshootermedia.com/">6ix Shooter Media</a>,
    Creative Commons license.<br/>
  </div>
</center>
</body>
</html>
"""

###############################################################################

_SIDEBAR_HEADER = """
    <div id="sidebar">
      <h1>Contents</h1>
      <div class="submenu">
        <a href="#">Top</a>
        <a href="#intro">Overview</a>
        <a href="#samples">Reference Panel</a>
"""

_SIDEBAR_NUCLEAR = """
        <a href="#admixture">Admixture Estimates</a>
        <a href="#pca">PCA Plots</a>
        <a href="#treemix">Treemix Analyses</a>
"""

_SIDEBAR_MITO = """
        <a href="#mito_phylo">MT Phylogeny</a>
"""

_SIDEBAR_FOOTER = """
        <a href="#references">References</a>
      </div>
    </div>
"""



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix
import paleomix.resources

from paleomix.node import Node

import paleomix.common.fileutils as fileutils

import paleomix.tools.zonkey.parts.admixture as admixture

from paleomix.tools.zonkey.parts.report import AnalysisReport


class SummaryNode(Node):
    def __init__(self, config, dependencies=()):
        self._root = config.destination
        self._data = config.database
        self._samples = config.samples
        self._sample_keys = self._samples.keys()

        input_files = set()
        self._reports = {}
        for sample, info in self._samples.iteritems():
            report = AnalysisReport(config=config,
                                    root=os.path.join(self._root, sample),
                                    has_nuc="Nuc" in info["Files"],
                                    has_mt="Mito" in info["Files"])

            input_files.update(report.input_files())
            self._reports[sample] = report

        Node.__init__(self,
                      input_files=input_files,
                      output_files=(os.path.join(self._root, "summary.html"),
                                    os.path.join(self._root, "summary.css")),
                      dependencies=dependencies)

    def _run(self, _config, temp):
        with open(os.path.join(temp, "summary.html"), "w") as output_handle:
            menu_entries = self._build_sidemenu()
            html_header = _HTML_HEADER.format(Version=paleomix.__version__,
                                              MenuEntries=menu_entries)
            output_handle.write(html_header)

            for sample in sorted(self._samples):
                self._write_sample_overview(output_handle, sample)

            output_handle.write(_HTML_FOOTER)

    def _teardown(self, config, temp):
        fileutils.make_dirs(self._root)

        fileutils.move_file(os.path.join(temp, "summary.html"),
                            os.path.join(self._root, "summary.html"))

        css_path = paleomix.resources.report("zonkey", "report.css")
        fileutils.copy_file(css_path, os.path.join(self._root, "summary.css"))

    def _build_sidemenu(self):
        lines = []
        for sample in sorted(self._samples):
            lines.append('        <a href="#sample_{0}">{0}</a>'.format(sample))
        return "\n".join(lines)

    def _write_sample_overview(self, handle, sample):
        info = self._samples[sample]
        sample_idx = self._sample_keys.index(sample)

        admixture = "&nbsp;"
        if "Nuc" in info["Files"]:
            admixture = self._read_admixture_results(sample)

        handle.write(_SAMPLE_HEADER.format(sample=sample,
                                           admixture=admixture))

        handle.write("""    <ul id="tabs">""")
        handle.write(_SAMPLE_TAB_SELECTED.format(sample_idx=sample_idx,
                                                 page=1, title="Overview"))
        handle.write(_SAMPLE_TAB.format(sample_idx=sample_idx,
                                        page=2, title="Sample Data"))

        if "Nuc" in info["Files"]:
            handle.write(_SAMPLE_TAB.format(sample_idx=sample_idx,
                                            page=3, title="PCA Plots"))
            handle.write(_SAMPLE_TAB.format(sample_idx=sample_idx,
                                            page=4, title="Treemix Plots"))

        if "Mito" in info["Files"]:
            handle.write(_SAMPLE_TAB.format(sample_idx=sample_idx,
                                            page=5,
                                            title="Mitochondrial Phylogeny"))

        handle.write("""    </ul>""")

        self._write_overview(handle, sample, sample_idx)

        if "Nuc" in info["Files"]:
            handle.write(_SAMPLE_PAGE_PCA.format(sample=sample,
                                                 sample_idx=sample_idx))
            handle.write(_SAMPLE_PAGE_TREEMIX.format(sample=sample,
                                                     sample_idx=sample_idx))

        if "Mito" in info["Files"]:
            handle.write(_SAMPLE_PAGE_MITO_PHYLO.format(sample=sample,
                                                        sample_idx=sample_idx))

        handle.write(_SAMPLE_FOOTER)

    def _write_overview(self, output_handle, sample, sample_idx,
                        cutoff=admixture.CUTOFF):
        info = self._samples[sample]
        report = self._reports[sample]

        n_tests = 4
        n_pos_tests = 0

        if "Nuc" in info["Files"]:

            for postfix in ('incl_ts', 'excl_ts'):
                admix_root = os.path.join(self._root, sample, "admixture")

                for k_groups in (2, 3):
                    filename = os.path.join(admix_root, "%s.%i.Q" % (postfix,
                                                                     k_groups))

                    result = admixture.read_admixture_results(filename,
                                                              self._data,
                                                              k_groups)

                    if sum(value >= cutoff for _, value in result) > 1:
                        n_pos_tests += 1

            if n_pos_tests:
                tmpl = _SAMPLE_OVERVIEW_INCL_NUCL_POSITIVE
            else:
                tmpl = _SAMPLE_OVERVIEW_INCL_NUCL_NEGATIVE
        else:
            tmpl = _SAMPLE_OVERVIEW_EXCL_NUCL

        output_handle.write(tmpl.format(sample_idx=sample_idx,
                                        n_tests=n_tests,
                                        n_pos_tests=n_pos_tests))

        output_handle.write(_SAMPLE_DATA_HEADER.format(sample_idx=sample_idx))

        if "Nuc" in info["Files"]:
            summary = report.snp_summary()
            output_handle.write(_SAMPLE_DATA_NUCLEAR % summary)

        if "Mito" in info["Files"]:
            summary = report.mito_summary()
            output_handle.write(_SAMPLE_DATA_MITOCHONDRIA % summary)

        output_handle.write(_SAMPLE_DATA_FOOTER)

    def _read_admixture_results(self, sample, cutoff=admixture.CUTOFF):
        lines = []
        lines.append('            <table summary="Admixture overview for sample {}" style="width:125px;">'.format(sample.replace('"', "")))
        lines.append("              <tr>")
        for postfix in ('incl_ts', 'excl_ts'):
            admix_root = os.path.join(self._root, sample, "admixture")
            for k_groups in (2, 3):
                filename = os.path.join(admix_root, "%s.%i.Q" % (postfix, k_groups))

                lines.append("                <td>")
                try:
                    ancestral_groups = admixture.read_admixture_results(filename, self._data, k_groups, cutoff)
                    lines.extend(self._build_admixture_figure([value for _, value in ancestral_groups]))
                except admixture.AdmixtureError:
                    lines.append('                  <div style="height:100px;background:gray"></div>')
                lines.append("                </td>")
        lines.append("              </tr>")
        lines.append("            </table>\n")

        return "\n".join(lines)

    @classmethod
    def _build_admixture_figure(cls, fractions, max_height=100):
        lines = []
        if len(fractions) == 2:
            colors = ("red", "green")
        elif len(fractions) == 3:
            colors = ("red", "green", "blue")
        else:
            raise RuntimeError("Unexpected number of fractions: %r"
                               % (fractions,))

        values = [value for value in fractions if value >= admixture.CUTOFF]

        tmpl = ' ' * 18 + '<div style="height:{}px;background:{}"></div>'
        for index, value in enumerate(values):
            height = round((float(value) / sum(values)) * max_height)
            lines.append(tmpl.format(height, colors[index]))

        return lines


###############################################################################

_TS_LABELS = {
    "incl_ts": "Including transitions",
    "excl_ts": "Excluding transitions",
}


###############################################################################

_HTML_HEADER = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
    <title>PALEOMIX Zonkey v{Version}</title>
    <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
    <link href="summary.css" rel="stylesheet" type="text/css" />

    <script type="text/javascript">
        function selectTab(sample, page) {{
            for (var i = 1; i <= 5; ++i) {{
                elem_id = "sample" + sample.toString() + ".link" + i.toString()

                var elem = document.getElementById(elem_id);
                if (elem) {{
                    elem.className = (i == page) ? 'selected' : '';
                }}

                elem_id = "sample" + sample.toString() + ".page" + i.toString()
                var elem = document.getElementById(elem_id);
                if (elem) {{
                    if (i == page) {{
                        if (page == 1) {{
                            elem.className = 'tabContent small';
                        }} else {{
                            elem.className = 'tabContent';
                        }}
                    }} else {{
                        elem.className = 'tabContent hide';
                    }}
                }}
            }}
        }}
    </script>
</head>
<body>
  <a name="top" id="top"></a>
  <center>
    <div id="header">
      <h1>PALEOMIX Zonkey {Version}</h1>
      <h2>A pipeline for detection of F1 hybrids in equids.</h2>
    </div>

    <div id="content" style="width:1050px;">
      <p class="introduction">
        Schubert M, Ermini L, Sarkissian CD, J&oacute;nsson H, Ginolhac A,
        Schaefer R, Martin MD, Fern&aacute;ndez R, Kircher M, McCue M,
        Willerslev E, and Orlando L. "<strong>Characterization of ancient and
        modern genomes by SNP detection and phylogenomic and metagenomic
        analysis using PALEOMIX</strong>". Nat Protoc. 2014 May;9(5):1056-82.
        doi:<a href="https://doi.org/10.1038/nprot.2014.063">
            10.1038/nprot.2014.063
        </a>.
        Epub 2014 Apr 10. PubMed PMID:
        <a href="http://www.ncbi.nlm.nih.gov/pubmed/24722405">24722405</a>.
      </p>
      <div id="sidebar">
        <h1>Contents</h1>
        <div class="submenu">
          <a href="#">Top</a>
          {MenuEntries}
        </div>
      </div>
      <div id="mainbar">
        <h1><a name="introduction" id="introduction"></a>Introduction</h1>
        <p>
          TODO
        </p>
        <h1><a name="samples" id="samples"></a>Sample Overview</h1>
"""


_SAMPLE_HEADER = """
    <h2 style="margin: 0px; padding: 0px;">
      <a name="sample_{sample}" id="sample_{sample}"
         href="{sample}/report.html">{sample}</a>
    </h2>

    <table style="margin: 0px; padding: 0px;">
      <tr>
        <td style="width:150px; vertical-align:top; padding-top:20px;">
{admixture}
        </td>
        <td>
"""

_SAMPLE_FOOTER = """
        </td>
      </tr>
    </table>
"""

_SAMPLE_TAB_SELECTED = """
      <li>
        <a id="sample{sample_idx}.link{page}"
           href="javascript:selectTab({sample_idx}, {page});"
           class="selected">{title}</a>
      </li>
"""

_SAMPLE_TAB = """
      <li>
        <a id="sample{sample_idx}.link{page}"
           href="javascript:selectTab({sample_idx}, {page});">{title}</a>
      </li>
"""


_SAMPLE_OVERVIEW_INCL_NUCL_POSITIVE = """
    <div class="tabContent small" id="sample{sample_idx}.page1">
      <p>
        Possible admixture detected in {n_pos_tests} of {n_tests} tests.
      </p>
    </div>
"""


_SAMPLE_OVERVIEW_INCL_NUCL_NEGATIVE = """
    <div class="tabContent small" id="sample{sample_idx}.page1">
      <p>
        No admixture detected in nuclear genome.
      </p>
    </div>
"""


_SAMPLE_OVERVIEW_EXCL_NUCL = """
    <div class="tabContent small" id="sample{sample_idx}.page1">
      <p>
        Nuclear BAM not provided; no admixture were tests perfomed.
      </p>
    </div>
"""


_SAMPLE_DATA_HEADER = """
    <div class="tabContent hide" id="sample{sample_idx}.page2">
      <ul>
"""


_SAMPLE_DATA_NUCLEAR = """
      <strong>Nuclear report:</strong>
      <ul>
        <li>BAM file: <em>%(filename)s</em></li>
        <li>Number of SNPs used (incl. transitions):
            <em>%(n_sites_incl_ts)s</em></li>
        <li>Number of SNPs used (excl. transitions):
            <em>%(n_sites_excl_ts)s</em></li>
      </ul>

      <br>
"""


_SAMPLE_DATA_MITOCHONDRIA = """
      <strong>Mitochondrial report:</strong>
      <ul>
        <li>BAM file: <em>%(filename)s</em></li>
        <li>Percentage of sites covered: <em>%(covered_pct)s</em></li>
        <li>Mean coverage per site: <em>%(mean_coverage)s</em></li>
      </ul>
"""


_SAMPLE_DATA_FOOTER = """
      </ul>
    </div>
"""


_SAMPLE_PAGE_PCA = """
    <div class="tabContent hide" id="sample{sample_idx}.page3">
      <table>
        <tr style="background-color:#f1f0ee">
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Including transitions</strong>

              <div class="image">
                <a href="{sample}/figures/pca/incl_ts.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/pca/incl_ts.png"
                       alt="PCA plot for {sample}, including transitions."/>
                </a>
              </div>
            </div>
          </td>
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Excluding transitions</strong>
              <div class="image">
                <a href="{sample}/figures/pca/excl_ts.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/pca/excl_ts.png"
                       alt="PCA plot for {sample}, excluding transitions."/>
                </a>
              </div>
            </div>
          </td>
        </tr>
      </table>
    </div>
"""


_SAMPLE_PAGE_TREEMIX = """
    <div class="tabContent hide" id="sample{sample_idx}.page4">
      <table>
        <tr style="background-color:#f1f0ee">
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Including transitions</strong>
              <div class="image">
                <a href="{sample}/figures/treemix/incl_ts_1_tree.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/treemix/incl_ts_1_tree.png"
                       alt="Treemix plot for {sample}, including transitions,
                            with one migration edge."/>
                </a>
              </div>
            </div>
          </td>
          <td style="text-align: center;">
            <div class="thumbnail">
              <strong>Excluding transitions</strong>
              <div class="image">
                <a href="{sample}/figures/treemix/excl_ts_1_tree.pdf">
                  <img style="width:13em"
                       src="{sample}/figures/treemix/excl_ts_1_tree.png"
                       alt="Treemix plot for {sample}, excluding transitions,
                            with one migration edge."></img>
                </a>
              </div>
            </div>
          </td>
        </tr>
      </table>
    </div>
"""


_SAMPLE_PAGE_MITO_PHYLO = """
    <div class="tabContent hide" id="sample{sample_idx}.page5">
      <div class="thumbnail" style="text-align: center;">
        <div class="image">
          <a href="{sample}/figures/mitochondria/mito_phylo.pdf">
            <img style="width:15em"
                 src="{sample}/figures/mitochondria/mito_phylo.png"
                 alt="Mitochondrial phylogeny for {sample}."/>
          </a>
        </div>
      </div>
    </div>
"""


_HTML_FOOTER = """
      </div>
    </div>
    <div id="footer">
        This report is based on the PLAIN 1.0 design by
        <a href="http://www.sixshootermedia.com/">6ix Shooter Media</a>,
        Creative Commons license.<br/>
    </div>
  </center>
</body>
</html>
"""



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import StringIO

import nose
from nose.tools import assert_equal
from paleomix.common.testing import with_temp_folder

from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import ParallelCmds, SequentialCmds
from paleomix.atomiccmd.pprint import pprint, pformat, _pformat_list


###############################################################################
###############################################################################
# pprint

def test_pprint__simple():
    obj = StringIO.StringIO()
    cmd = AtomicCmd(("touch", "something"))
    pprint(cmd, out=obj)
    assert_equal(obj.getvalue(), ("<Command = ['touch', 'something']\n"
                                  " STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
                                  " STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'>\n") % (id(cmd), id(cmd)))


###############################################################################
###############################################################################
# INFILE

def test_pformat__simple():
    cmd = AtomicCmd(("touch", "something"))
    assert_equal(pformat(cmd), ("<Command = ['touch', 'something']\n"
                                " STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'>") % (id(cmd), id(cmd)))


@with_temp_folder
def test_pformat__simple__running(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    assert_equal(pformat(cmd), ("<Command = ['sleep', '10']\n"
                                " Status  = Running ...\n"
                                " STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
                                " STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
                                " CWD     = '{cwd}'>").format(id=id(cmd),
                                                              cwd=os.getcwd(),
                                                              temp_dir=temp_folder))
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_pformat__simple__running__set_cwd(temp_folder):
    cmd = AtomicCmd(("sleep", "10"), set_cwd=True)
    cmd.run(temp_folder)
    assert_equal(pformat(cmd), ("<Command = ['sleep', '10']\n"
                                " Status  = Running ...\n"
                                " STDOUT* = 'pipe_sleep_{id}.stdout'\n"
                                " STDERR* = 'pipe_sleep_{id}.stderr'\n"
                                " CWD     = '{temp_dir}'>").format(id=id(cmd),
                                                                   temp_dir=temp_folder))
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_pformat__simple__done(temp_folder):
    cmd = AtomicCmd("true")
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert_equal(pformat(cmd), ("<Command = ['true']\n"
                                " Status  = Exited with return-code 0\n"
                                " STDOUT* = '{temp_dir}/pipe_true_{id}.stdout'\n"
                                " STDERR* = '{temp_dir}/pipe_true_{id}.stderr'\n"
                                " CWD     = '{cwd}'>").format(id=id(cmd),
                                                              cwd=os.getcwd(),
                                                              temp_dir=temp_folder))


@with_temp_folder
def test_pformat__simple__done__before_join(temp_folder):
    cmd = AtomicCmd("true")
    cmd.run(temp_folder)
    cmd._proc.wait()
    assert_equal(pformat(cmd), ("<Command = ['true']\n"
                                " Status  = Exited with return-code 0\n"
                                " STDOUT* = '{temp_dir}/pipe_true_{id}.stdout'\n"
                                " STDERR* = '{temp_dir}/pipe_true_{id}.stderr'\n"
                                " CWD     = '{cwd}'>").format(id=id(cmd),
                                                              cwd=os.getcwd(),
                                                              temp_dir=temp_folder))
    assert_equal(cmd.join(), [0])


@with_temp_folder
def test_pformat__simple__done__set_cwd(temp_folder):
    cmd = AtomicCmd("true", set_cwd=True)
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert_equal(pformat(cmd), ("<Command = ['true']\n"
                                " Status  = Exited with return-code 0\n"
                                " STDOUT* = 'pipe_true_{id}.stdout'\n"
                                " STDERR* = 'pipe_true_{id}.stderr'\n"
                                " CWD     = '{temp_dir}'>").format(id=id(cmd),
                                                                   temp_dir=temp_folder))


@with_temp_folder
def test_pformat__simple__killed(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    cmd.terminate()
    assert_equal(cmd.join(), ["SIGTERM"])
    assert_equal(pformat(cmd), ("<Command = ['sleep', '10']\n"
                                " Status  = Terminated with signal SIGTERM\n"
                                " STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
                                " STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
                                " CWD     = '{cwd}'>").format(id=id(cmd),
                                                              temp_dir=temp_folder,
                                                              cwd=os.getcwd()))


###############################################################################
###############################################################################
# INFILE

def test_pformat__atomiccmd__simple_with_infile():
    cmd = AtomicCmd(("cat", "%(IN_SOMETHING)s"), IN_SOMETHING="/etc/fstab")
    assert_equal(pformat(cmd), ("<Command = ['cat', '/etc/fstab']\n"
                                " STDOUT* = '${TEMP_DIR}/pipe_cat_%i.stdout'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_cat_%i.stderr'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_infile__set_cwd():
    cmd = AtomicCmd(("cat", "%(IN_SOMETHING)s"),
                    IN_SOMETHING="/etc/fstab",
                    set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['cat', '/etc/fstab']\n"
                                " STDOUT* = 'pipe_cat_%i.stdout'\n"
                                " STDERR* = 'pipe_cat_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_infile():
    cmd = AtomicCmd(("cat", "%(TEMP_IN_FILE)s"),
                    TEMP_IN_FILE="infile.txt")
    assert_equal(pformat(cmd), ("<Command = ['cat', '${TEMP_DIR}/infile.txt']\n"
                                " STDOUT* = '${TEMP_DIR}/pipe_cat_%i.stdout'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_cat_%i.stderr'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_infile__set_cwd():
    cmd = AtomicCmd(("zcat", "%(TEMP_IN_FILE)s"),
                    TEMP_IN_FILE="infile.gz",
                    set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['zcat', 'infile.gz']\n"
                                " STDOUT* = 'pipe_zcat_%i.stdout'\n"
                                " STDERR* = 'pipe_zcat_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd), id(cmd)))


###############################################################################
###############################################################################
# OUTFILE

def test_pformat__atomiccmd__simple_with_outfile():
    cmd = AtomicCmd(("touch", "%(OUT_RC)s"), OUT_RC="/etc/bashrc")
    assert_equal(pformat(cmd), ("<Command = ['touch', '${TEMP_DIR}/bashrc']\n"
                                " STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_outfile__set_cwd():
    cmd = AtomicCmd(("touch", "%(OUT_RC)s"), OUT_RC="/etc/bashrc", set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['touch', 'bashrc']\n"
                                " STDOUT* = 'pipe_touch_%i.stdout'\n"
                                " STDERR* = 'pipe_touch_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_outfile():
    cmd = AtomicCmd(("touch", "%(TEMP_OUT_RC)s"), TEMP_OUT_RC="bashrc")
    assert_equal(pformat(cmd), ("<Command = ['touch', '${TEMP_DIR}/bashrc']\n"
                                " STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_outfile__set_cwd():
    cmd = AtomicCmd(("touch", "%(TEMP_OUT_RC)s"), TEMP_OUT_RC="bashrc", set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['touch', 'bashrc']\n"
                                " STDOUT* = 'pipe_touch_%i.stdout'\n"
                                " STDERR* = 'pipe_touch_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd), id(cmd)))


###############################################################################
###############################################################################
# STDIN

def test_pformat__atomiccmd__simple_with_stdin():
    cmd = AtomicCmd("gzip", IN_STDIN="/etc/fstab")
    assert_equal(pformat(cmd), ("<Command = ['gzip']\n"
                                " STDIN   = '/etc/fstab'\n"
                                " STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_stdin__set_cwd():
    cmd = AtomicCmd("gzip", IN_STDIN="/etc/fstab", set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['gzip']\n"
                                " STDIN   = '/etc/fstab'\n"
                                " STDOUT* = 'pipe_gzip_%i.stdout'\n"
                                " STDERR* = 'pipe_gzip_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_stdin():
    cmd = AtomicCmd("gzip", TEMP_IN_STDIN="stabstabstab")
    assert_equal(pformat(cmd), ("<Command = ['gzip']\n"
                                " STDIN*  = '${TEMP_DIR}/stabstabstab'\n"
                                " STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_temp_stdin__set_cwd():
    cmd = AtomicCmd("gzip", TEMP_IN_STDIN="stabstabstab", set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['gzip']\n"
                                " STDIN*  = 'stabstabstab'\n"
                                " STDOUT* = 'pipe_gzip_%i.stdout'\n"
                                " STDERR* = 'pipe_gzip_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd), id(cmd)))


def test_pformat__atomiccmd__simple_with_stdin__cmd():
    cmd_1 = AtomicCmd("gzip", OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
    assert_equal(pformat(cmd_2), ("<Command = ['gzip']\n"
                                  " STDIN   = <PIPE>\n"
                                  " STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
                                  " STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'>") % (id(cmd_2), id(cmd_2)))


###############################################################################
###############################################################################
# STDOUT

def test_pformat__atomiccmd__simple_with_stdout():
    cmd = AtomicCmd(("echo", "Water. Water."), OUT_STDOUT="/dev/ls")
    assert_equal(pformat(cmd), ("<Command = ['echo', 'Water. Water.']\n"
                                " STDOUT  = '${TEMP_DIR}/ls'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'>") % (id(cmd),))


def test_pformat__atomiccmd__simple_with_stdout__set_cwd():
    cmd = AtomicCmd(("echo", "*pant*. *pant*."), OUT_STDOUT="/dev/barf", set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['echo', '*pant*. *pant*.']\n"
                                " STDOUT  = 'barf'\n"
                                " STDERR* = 'pipe_echo_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd),))


def test_pformat__atomiccmd__simple_with_temp_stdout():
    cmd = AtomicCmd(("echo", "Oil. Oil."), TEMP_OUT_STDOUT="dm")
    assert_equal(pformat(cmd), ("<Command = ['echo', 'Oil. Oil.']\n"
                                " STDOUT* = '${TEMP_DIR}/dm'\n"
                                " STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'>") % (id(cmd),))


def test_pformat__atomiccmd__simple_with_temp_stdout__set_cwd():
    cmd = AtomicCmd(("echo", "Room service. Room service."),
                    TEMP_OUT_STDOUT="pv",
                    set_cwd=True)
    assert_equal(pformat(cmd), ("<Command = ['echo', 'Room service. Room service.']\n"
                                " STDOUT* = 'pv'\n"
                                " STDERR* = 'pipe_echo_%i.stderr'\n"
                                " CWD     = '${TEMP_DIR}'>") % (id(cmd),))


def test_pformat__atomiccmd__simple_with_stdout_pipe():
    cmd = AtomicCmd(("echo", "!"), OUT_STDOUT=AtomicCmd.PIPE)
    assert_equal(pformat(cmd), ("<Command = ['echo', '!']\n"
                                " STDOUT  = <PIPE>\n"
                                " STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'>") % (id(cmd),))


###############################################################################
###############################################################################
# ParallelCmds

def test_pformat__sets__simple():
    def _do_test_pformat__sets__simple(cls, description):
        cmd_1 = AtomicCmd(("echo", "foo"), OUT_STDOUT=AtomicCmd.PIPE)
        cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
        cmd = cls((cmd_1, cmd_2))
        assert_equal(pformat(cmd),
                     ("<{description}:\n"
                      "  - <00> Command = ['echo', 'foo']\n"
                      "         STDOUT  = <01>\n"
                      "         STDERR* = '${{TEMP_DIR}}/pipe_echo_{cmd_1_id}.stderr'\n"
                      "  - <01> Command = ['gzip']\n"
                      "         STDIN   = <00>\n"
                      "         STDOUT* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stdout'\n"
                      "         STDERR* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stderr'>")
                     .format(description=description,
                             cmd_1_id=id(cmd_1),
                             cmd_2_id=id(cmd_2)))
    yield _do_test_pformat__sets__simple, ParallelCmds, "Parallel commands"
    yield _do_test_pformat__sets__simple, SequentialCmds, "Sequential commands"


def test_pformat__sets__nested():
    cmd_1 = AtomicCmd(("echo", "foo"), OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
    cmd_3 = AtomicCmd("sha1sum")
    set_1 = ParallelCmds((cmd_1, cmd_2))
    set_2 = SequentialCmds((set_1, cmd_3))
    assert_equal(pformat(set_2),
                 ("<Sequential commands:\n"
                  "  - Parallel commands:\n"
                  "      - <00> Command = ['echo', 'foo']\n"
                  "             STDOUT  = <01>\n"
                  "             STDERR* = '${{TEMP_DIR}}/pipe_echo_{cmd_1_id}.stderr'\n"
                  "      - <01> Command = ['gzip']\n"
                  "             STDIN   = <00>\n"
                  "             STDOUT* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stdout'\n"
                  "             STDERR* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stderr'\n"
                  "  - <02> Command = ['sha1sum']\n"
                  "         STDOUT* = '${{TEMP_DIR}}/pipe_sha1sum_{cmd_3_id}.stdout'\n"
                  "         STDERR* = '${{TEMP_DIR}}/pipe_sha1sum_{cmd_3_id}.stderr'>")
                 .format(cmd_1_id=id(cmd_1),
                         cmd_2_id=id(cmd_2),
                         cmd_3_id=id(cmd_3)))


###############################################################################
###############################################################################
# Bad input

def test_pformat__bad_input():
    @nose.tools.raises(TypeError)
    def _do_test_pformat__bad_input(value):
        pformat(value)
    yield _do_test_pformat__bad_input, 1
    yield _do_test_pformat__bad_input, {}
    yield _do_test_pformat__bad_input, ""


###############################################################################
###############################################################################
# _pformat_list

def test_pformat_list__empty():
    assert_equal(_pformat_list([]), "[]")


def test_pformat_list__single():
    assert_equal(_pformat_list([3]), "[3]")


def test_pformat_list__multiple():
    assert_equal(_pformat_list([3, 2, 1]), "[3, 2, 1]")


def test_pformat_list__wrapped():
    assert_equal(_pformat_list([3, 2, 1], width=1), "[3,\n 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=2), "[3,\n 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=3), "[3,\n 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=4), "[3,\n 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=5), "[3,\n 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=6), "[3, 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=7), "[3, 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=8), "[3, 2,\n 1]")
    assert_equal(_pformat_list([3, 2, 1], width=9), "[3, 2, 1]")
    assert_equal(_pformat_list([3, 2, 1], width=10), "[3, 2, 1]")



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os
import sys
import time

import paleomix.logger
import paleomix.resources
import paleomix.tools.phylo_pipeline.mkfile as mkfile
import paleomix.ui
import paleomix.yaml

from paleomix.pipeline import Pypeline
from paleomix.common.console import print_err
from paleomix.tools.phylo_pipeline.makefile import \
    MakefileError, \
    read_makefiles
from paleomix.tools.phylo_pipeline.config import \
    ConfigError, \
    parse_config, \
    select_commands


def main(argv):
    try:
        config, args = parse_config(argv)
    except ConfigError, error:
        print_err(error)
        return 1

    if not args or ("help" in args):
        return 0
    elif args[0] in ("example", "examples"):
        if paleomix.resources.copy_example("phylo_pipeline", argv[1:]):
            return 1

        # Update interpreter to match the one currently in use;
        # this is required since we may be running from a virtual env
        filename = os.path.join(argv[1],
                                'phylo_pipeline',
                                'synthesize_reads.py')

        with open(filename) as handle:
            header, lines = handle.read().split('\n', 1)

        with open(filename, 'w') as handle:
            handle.write('#!%s\n' % (os.path.abspath(sys.executable, )))
            handle.write(lines)

        return 0
    elif (len(args) < 2) and ("mkfile" not in args and "makefile" not in args):
        print_err("\nPlease specify at least one makefile!")
        return 1

    commands = select_commands(args.pop(0))
    if any((cmd in ("makefile", "mkfile")) for (cmd, _) in commands):
        return mkfile.main(args[1:])

    if not os.path.exists(config.temp_root):
        try:
            os.makedirs(config.temp_root)
        except OSError, error:
            print_err("ERROR: Could not create temp root:\n\t%s" % (error,))
            return 1

    if not os.access(config.temp_root, os.R_OK | os.W_OK | os.X_OK):
        print_err("ERROR: Insufficient permissions for temp root: '%s'"
                  % (config.temp_root,))
        return 1

    # Init worker-threads before reading in any more data
    pipeline = Pypeline(config)

    try:
        makefiles = read_makefiles(config, args, commands)
    except (MakefileError, paleomix.yaml.YAMLError, IOError), error:
        print_err("Error reading makefiles:",
                  "\n  %s:\n   " % (error.__class__.__name__,),
                  "\n    ".join(str(error).split("\n")))
        return 1

    logfile_template = time.strftime("phylo_pipeline.%Y%m%d_%H%M%S_%%02i.log")
    paleomix.logger.initialize(config, logfile_template)
    logger = logging.getLogger(__name__)

    for (command_key, command_func) in commands:
        logger.info("Building %s pipeline ...", command_key)
        command_func(pipeline, config, makefiles)

    for makefile in makefiles:
        if "Nodes" in makefile:
            pipeline.add_nodes(makefile["Nodes"])

    if config.list_output_files:
        logger.info("Printing output files ...")
        pipeline.print_output_files()
        return 0
    elif config.list_executables:
        logger.info("Printing required executables ...")
        pipeline.print_required_executables()
        return 0
    elif config.dot_file:
        logger.info("Writing dependency graph to %r ...", config.dot_file)
        if not pipeline.to_dot(config.dot_file):
            return 1
        return 0

    if not pipeline.run(max_threads=config.max_threads,
                        dry_run=config.dry_run,
                        progress_ui=config.progress_ui):
        return 1
    return 0



#!/usr/bin/python
#
# Copyright (c) 2015 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import \
    MetaNode

from paleomix.tools.cnv_pipeline.steps.nodes import \
    MrFastIndexNode, \
    SplitFASTANode, \
    FASTAKMerNode, \
    MrFastMapNode, \
    CountHitsNode, \
    MergeBedsNode, \
    MaskBedNode, \
    CatFASTANode, \
    MrCNVRPrepNode

from paleomix.common.fileutils import \
    swap_ext


def build_map_kmer_nodes(prefix, reads, indices, dependencies=()):
    kmer_fpaths = []
    kmer_mappings = []

    for contig_fpath, indice_node in indices.iteritems():
        contig_name = os.path.basename(contig_fpath).split(".", 1)[0]

        output_prefix = "%s_%s" % (prefix, contig_name)
        map_node = MrFastMapNode(reads_fpath=reads,
                                 fasta_fpath=contig_fpath,
                                 output_prefix=output_prefix,
                                 dependencies=dependencies + (indice_node,))
        kmer_mappings.append(map_node)
        kmer_fpaths.extend(map_node.output_files)

    output_table = os.path.join(prefix + ".table")
    return CountHitsNode(input_files=kmer_fpaths,
                         output_file=output_table,
                         dependencies=kmer_mappings)


def build_prefix_pipelines(config, makefile):
    nodes = []
    for (key, prefix) in makefile["Prefixes"].iteritems():
        output_dir = os.path.join(config.destination, "prefixes", key)

        # 1. Split FASTA into a FASTA per contig
        node = SplitFASTANode(fasta_fpath=prefix["Path"],
                              output_dir=os.path.join(output_dir, "raw"))

        # 2. Index each contig using MrFAST
        indices = {}
        for contig_fpath, _ in node.contigs:
            indices[contig_fpath] = MrFastIndexNode(contig_fpath,
                                                    dependencies=(node,))

        # 3. Build KMers from each contig
        kmer_nodes = {}
        for contig_fpath, size in node.contigs:
            kmer_dir = os.path.splitext(contig_fpath)[0]
            kmer_nodes[contig_fpath] = FASTAKMerNode(fasta_fpath=contig_fpath,
                                                     output_dir=kmer_dir,
                                                     contigs={contig_fpath: size},
                                                     dependencies=(node,))

        # 4. Map each set of MKers against each contig
        kmer_hits_nodes = {}
        for contig_fpath, kmer_node in kmer_nodes.iteritems():
            mapping_nodes = []
            for kmer_fpath in kmer_node.output_files:
                contig_dirname, contig_fname = os.path.split(kmer_fpath)

                prefix = os.path.join(contig_dirname,
                                      contig_fname.split(".")[0])

                kmer_node = build_map_kmer_nodes(prefix=prefix,
                                                 reads=kmer_fpath,
                                                 indices=indices,
                                                 dependencies=(kmer_node,))

                mapping_nodes.append(kmer_node)
            kmer_hits_nodes[contig_fpath] = mapping_nodes

        # 5. Build masking beds from hits
        mask_nodes = {}
        for contig_fpath, size in node.contigs:
            kmer_files = set()
            for kmer_node in kmer_hits_nodes[contig_fpath]:
                kmer_files.update(kmer_node.output_files)

            contig_name = os.path.basename(contig_fpath).split(".", 1)[0]
            bed_fpaths = _get_bed_files(config.mask_dir)
            bed_node = MergeBedsNode(hits_files=kmer_files,
                                     bed_files=bed_fpaths,
                                     output_file=swap_ext(contig_fpath, ".bed"),
                                     contig=contig_name,
                                     dependencies=kmer_hits_nodes[contig_fpath])

            mask_nodes[contig_fpath] = bed_node

        # 6. Build and index masked FASTAs for MrFAST
        for contig_fpath, _ in node.contigs:
            bed_node = mask_nodes[contig_fpath]
            bed_fpath, = bed_node.output_files

            fasta_fpath = os.path.join(output_dir, "mrfast",
                                       os.path.basename(contig_fpath))
            mask_node = MaskBedNode(fasta_fpath=contig_fpath,
                                    bed_fpath=bed_fpath,
                                    output_file=fasta_fpath,
                                    dependencies=(bed_node,))

            index_node = MrFastIndexNode(fasta_fpath,
                                         dependencies=(mask_node,))

            nodes.append(index_node)

        # 7. Build, merge, and index masked FASTAs for MrCNVR
        masked_nodes = []
        masked_fpaths = []
        for contig_fpath, contig_size in node.contigs:
            bed_node = mask_nodes[contig_fpath]
            bed_fpath, = bed_node.output_files

            fasta_fpath = os.path.join(output_dir, "mrcnvr", "contigs",
                                       os.path.basename(contig_fpath))
            mask_node = MaskBedNode(fasta_fpath=contig_fpath,
                                    bed_fpath=bed_fpath,
                                    output_file=fasta_fpath,
                                    slop=16,
                                    contig_size=contig_size,
                                    dependencies=(bed_node,))

            masked_nodes.append(mask_node)
            masked_fpaths.append(fasta_fpath)

        fasta_fpath = os.path.join(output_dir, "mrcnvr", "merged.fasta")
        merge_node = CatFASTANode(fasta_fpaths=masked_fpaths,
                                  output_file=fasta_fpath,
                                  dependencies=masked_nodes)

        conf_node = MrCNVRPrepNode(fasta_fpath=fasta_fpath,
                                   output_file=swap_ext(fasta_fpath, ".conf"),
                                   dependencies=(merge_node,))

        nodes.append(conf_node)

    return MetaNode(description="Prefixes",
                    dependencies=nodes)


def build_reads_pipeline(config, makefile):
    records = _iterate_over_records(makefile)
    for target, sample, library, barcode, record in records:
        print target, sample, library, barcode

    return MetaNode()


def build_pipeline(config, makefile):
    return MetaNode(dependencies=(build_prefix_pipelines(config, makefile),
                                  build_reads_pipeline(config, makefile)))


def _get_bed_files(root):
    result = []
    # FIXME: better handling
    if os.path.isdir(root):
        for fname in os.listdir(root):
            result.append(os.path.join(root, fname))
    return result


# FIXME: Move to common module (also in bam_pipeline.makefile)
def _iterate_over_records(makefile):
    for (target, samples) in makefile["Targets"].items():
        for (sample, libraries) in samples.items():
            for (library, barcodes) in libraries.items():
                for (barcode, record) in barcodes.items():
                    yield target, sample, library, barcode, record



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MSchubert@snm.ku.dk>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import re
import glob

from paleomix.common.makefile import \
    MakefileError


def is_paired_end(template):
    """Returns true if a template contains a Pair component."""
    return (template.format(Pair=1) != template)


def collect_files(path, template):
    """

    """
    if is_paired_end(template):
        if _has_glob_magic(template):
            result = {"PE_1": _sorted_glob(template.format(Pair=1)),
                      "PE_2": _sorted_glob(template.format(Pair=2))}

            if not (result["PE_1"] or result["PE_2"]):
                _raise_missing_files("paired-end", path, template)
            elif len(result["PE_1"]) != len(result["PE_2"]):
                raise MakefileError("Unequal number of mate 1 and mate 2 "
                                    "files found at path %r; found %i mate 1 "
                                    "files, and %i mate 2 files; specified in "
                                    "makefile at %r. Please verify that the "
                                    "path is correct, and update the makefile!"
                                    % (template,
                                       len(result["PE_1"]),
                                       len(result["PE_2"]),
                                       " :: ".join(path)))
        else:
            result = {"PE_1": [template.format(Pair=1)],
                      "PE_2": [template.format(Pair=2)]}
    elif _has_glob_magic(template):
        result = {"SE": _sorted_glob(template)}
        if not result["SE"]:
            _raise_missing_files("single-end", path, template)
    else:
        result = {"SE": [template]}

    return result


def _has_glob_magic(filename):
    """Returns true if a path contains wildcards used by glob / fnmatch."""
    return _GLOB_MAGIC.search(filename) is not None


def _sorted_glob(tmpl):
    return list(sorted(glob.iglob(tmpl)))


def _raise_missing_files(description, path, template):
    raise MakefileError("No files found for %s reads using path %r; "
                        "specified in makefile at %r. Please verify that the "
                        "path is correct, and update the makefile!"
                        % (description, template, " :: ".join(path)))


_GLOB_MAGIC = re.compile('[*?[]')
